private import bogoyaml;
private import util;
private import std.stream;
private import std.string;

class YamlLangRunner {
    static YamlHolder yfalse;
    static YamlHolder ytrue;
    static this() {
        yfalse = new YamlScalar("");
        ytrue = new YamlScalar("1");
    }

    YamlHolder delegate()[char[]] bi;

    void addBuiltIn(char[] name, YamlHolder delegate() p) {
        bi[name] = p;
    }

    YamlHolder bih_get(char[] msg) {
        YamlHolder ret = invokeMsg(msg);
        if (!ret) {
            die(msg ~ ": not found");
        }
        return ret;
    }
    YamlHolder bih_getres(char[] msg) {
        return invoke(bih_get(msg));
    }
    char[] bih_getstr(char[] msg) {
        return bih_getres(msg).str;
    }

    void dumpContext() {
        for (int i = context.length-1; i >= 0; i--) {
            std.stream.stdout.writeLine(`- `);
            foreach (char[] k, YamlHolder v; context[i]) {
                std.stream.stdout.writeLine(" " ~ k ~ ":");
                if (v) v.dumpLocal(std.stream.stdout, 2);
                std.stream.stdout.writeLine("");
            }
        }
    }
    void die(char[] msg) {
        std.stream.stdout.writeLine(`--- ERROR CONTEXT`);
        dumpContext();
        .die(msg);
    }

    this(YamlHolder h) {
        context ~= h.assoc;
    }

    YamlHolder invokeMsg(char[] msg) {
        if (msg.length == 0) {
            die(msg ~ ": nil func");
        }

        if (msg[0] == '.') {
            char[] m = msg[1..msg.length];
            YamlHolder next;
            for (int i = context.length-1; i >= 0; i--) {
                if (m in context[i]) {
                    next = context[i][m];
                    break;
                }
            }
            if (!next) {
                die(msg ~ ": not found func");
            }

            YamlHolder ret = invoke(next);

            return ret;
        }
        else {
            die(msg ~ ": not func");
        }
    }

    YamlHolder invoke(YamlHolder next) {
        if (next.isArray) {
            YamlHolder[] procs = next.array;
            YamlHolder ret;
            foreach (YamlHolder h; procs) {
                ret = invoke(h);
            }
            return ret;
        }
        else if (next.isAssoc) {
            YamlHolder[char[]] assoc = next.assoc;

            if (assoc.length == 0) {
                return next;
            }

            char[] msg = assoc.keys[0];

            if (msg[0] == '.') {
                assert(assoc.keys.length == 1);

                context.length = context.length + 1;

                YamlHolder args = invoke(assoc[msg]);
                if (!args.isAssoc) {
                    die(msg ~ ": invalid argument");
                }

                YamlHolder ret;
                if (msg in bi) {
                    ret = bi[msg]();
                }
                else {
                    ret = invokeMsg(msg);
                }

                context.length = context.length - 1;

                return ret;
            }
            else {
                foreach (char[] k, YamlHolder h; assoc) {
                    if (h.isArray) {
                        context[context.length-1][k] = h;
                    }
                    else {
                        context[context.length-1][k] = invoke(h);
                    }
                }
                return next;
            }
        }
        else {
            char[] ret = next.str;
            if (ret.length > 0 && ret[0] == '.') {
                YamlHolder nmsg = invokeMsg(ret);
                return invoke(nmsg);
            }
            else {
                return next;
            }
        }
    }

    YamlHolder[char[]][] context;
}

